__author__ = "Lukas Leufen, Felix Kleinert"
__date__ = '2019-12-17'

import logging
import math
import os
import warnings
from typing import Dict, List, Tuple


import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import xarray as xr
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.patches as mpatches

from src import helpers
from src.helpers import TimeTracking, TimeTrackingWrapper
from src.data_handling.data_generator import DataGenerator

logging.getLogger('matplotlib').setLevel(logging.WARNING)


class AbstractPlotClass:

    def __init__(self, plot_folder, plot_name, resolution=500):
        self.plot_folder = plot_folder
        self.plot_name = plot_name
        self.resolution = resolution

    def _plot(self, *args):
        raise NotImplementedError

    def _save(self, **kwargs):
        """
        Standard save method to store plot locally. Name of and path to plot need to be set on initialisation
        """
        plot_name = os.path.join(os.path.abspath(self.plot_folder), f"{self.plot_name}.pdf")
        logging.debug(f"... save plot to {plot_name}")
        plt.savefig(plot_name, dpi=self.resolution, **kwargs)
        plt.close('all')


@TimeTrackingWrapper
class PlotMonthlySummary(AbstractPlotClass):
    """
    Show a monthly summary over all stations for each lead time ("ahead") as box and whiskers plot. The plot is saved
    in data_path with name monthly_summary_box_plot.pdf and 500dpi resolution.
    """
    def __init__(self, stations: List, data_path: str, name: str, target_var: str, window_lead_time: int = None,
                 plot_folder: str = "."):
        """
        Sets attributes and create plot
        :param stations: all stations to plot
        :param data_path: path, where the data is located
        :param name: full name of the local files with a % as placeholder for the station name
        :param target_var: display name of the target variable on plot's axis
        :param window_lead_time: lead time to plot, if window_lead_time is higher than the available lead time or not given
            the maximum lead time from data is used. (default None -> use maximum lead time from data).
        :param plot_folder: path to save the plot (default: current directory)
        """
        super().__init__(plot_folder, "monthly_summary_box_plot")
        self._data_path = data_path
        self._data_name = name
        self._data = self._prepare_data(stations)
        self._window_lead_time = self._get_window_lead_time(window_lead_time)
        self._plot(target_var)
        self._save()

    def _prepare_data(self, stations: List) -> xr.DataArray:
        """
        Pre-process data required to plot. For each station, load locally saved predictions, extract the CNN prediction
        and the observation and group them into monthly bins (no aggregation, only sorting them).
        :param stations: all stations to plot
        :return: The entire data set, flagged with the corresponding month.
        """
        forecasts = None
        for station in stations:
            logging.debug(f"... preprocess station {station}")
            file_name = os.path.join(self._data_path, self._data_name % station)
            data = xr.open_dataarray(file_name)

            data_cnn = data.sel(type="CNN").squeeze()
            if len(data_cnn.shape) > 1:
                data_cnn.coords["ahead"].values = [f"{days}d" for days in data_cnn.coords["ahead"].values]

            data_obs = data.sel(type="obs", ahead=1).squeeze()
            data_obs.coords["ahead"] = "obs"

            data_concat = xr.concat([data_obs, data_cnn], dim="ahead")
            data_concat = data_concat.drop("type")

            data_concat.index.values = data_concat.index.values.astype("datetime64[M]").astype(int) % 12 + 1
            data_concat = data_concat.clip(min=0)

            forecasts = xr.concat([forecasts, data_concat], 'index') if forecasts is not None else data_concat
        return forecasts

    def _get_window_lead_time(self, window_lead_time: int):
        """
        Extract the lead time from data and arguments. If window_lead_time is not given, extract this information from
        data itself by the number of ahead dimensions. If given, check if data supports the give length. If the number
        of ahead dimensions in data is lower than the given lead time, data's lead time is used.
        :param window_lead_time: lead time from arguments to validate
        :return: validated lead time, comes either from given argument or from data itself
        """
        ahead_steps = len(self._data.ahead)
        if window_lead_time is None:
            window_lead_time = ahead_steps
        return min(ahead_steps, window_lead_time)

    def _plot(self, target_var: str):
        """
        Main plot function that creates a monthly grouped box plot over all stations but with separate boxes for each
        lead time step.
        :param target_var: display name of the target variable on plot's axis
        """
        data = self._data.to_dataset(name='values').to_dask_dataframe()
        logging.debug("... start plotting")
        color_palette = [matplotlib.colors.cnames["green"]] + sns.color_palette("Blues_d", self._window_lead_time).as_hex()
        ax = sns.boxplot(x='index', y='values', hue='ahead', data=data.compute(), whis=1., palette=color_palette,
                         flierprops={'marker': '.', 'markersize': 1}, showmeans=True,
                         meanprops={'markersize': 1, 'markeredgecolor': 'k'})
        ax.set(xlabel='month', ylabel=f'{target_var}')
        plt.tight_layout()


@TimeTrackingWrapper
class PlotStationMap(AbstractPlotClass):
    """
    Plot geographical overview of all used stations as squares. Different data sets can be colorised by its key in the
    input dictionary generators. The key represents the color to plot on the map. Currently, there is only a white
    background, but this can be adjusted by loading locally stored topography data (not implemented yet). The plot is
    saved under plot_path with the name station_map.pdf
    """

    def __init__(self, generators: Dict, plot_folder: str = "."):
        """
        Sets attributes and create plot
        :param generators: dictionary with the plot color of each data set as key and the generator containing all stations
        as value.
        :param plot_folder: path to save the plot (default: current directory)
        """
        super().__init__(plot_folder, "station_map")
        self._ax = None
        self._plot(generators)
        self._save()

    def _draw_background(self):
        """
        Draw coastline, lakes, ocean, rivers and country borders as background on the map.
        """

        import cartopy.feature as cfeature
        self._ax.add_feature(cfeature.COASTLINE.with_scale("50m"), edgecolor='black')
        self._ax.add_feature(cfeature.LAKES.with_scale("50m"))
        self._ax.add_feature(cfeature.OCEAN.with_scale("50m"))
        self._ax.add_feature(cfeature.RIVERS.with_scale("50m"))
        self._ax.add_feature(cfeature.BORDERS.with_scale("50m"), facecolor='none', edgecolor='black')

    def _plot_stations(self, generators):
        """
        The actual plot function. Loops over all keys in generators dict and its containing stations and plots a square
        and the stations's position on the map regarding the given color.
        :param generators: dictionary with the plot color of each data set as key and the generator containing all
            stations as value.
        """

        import cartopy.crs as ccrs
        if generators is not None:
            for color, gen in generators.items():
                for k, v in enumerate(gen):
                    station_coords = gen.get_data_generator(k).meta.loc[['station_lon', 'station_lat']]
                    # station_names = gen.get_data_generator(k).meta.loc[['station_id']]
                    IDx, IDy = float(station_coords.loc['station_lon'].values), float(
                        station_coords.loc['station_lat'].values)
                    self._ax.plot(IDx, IDy, mfc=color, mec='k', marker='s', markersize=6, transform=ccrs.PlateCarree())

    def _plot(self, generators: Dict):
        """
        Main plot function to create the station map plot. Sets figure and calls all required sub-methods.
        :param generators: dictionary with the plot color of each data set as key and the generator containing all
            stations as value.
        """

        import cartopy.crs as ccrs
        fig = plt.figure(figsize=(10, 5))
        self._ax = fig.add_subplot(1, 1, 1, projection=ccrs.PlateCarree())
        self._ax.set_extent([0, 20, 42, 58], crs=ccrs.PlateCarree())
        self._draw_background()
        self._plot_stations(generators)


@TimeTrackingWrapper
class PlotConditionalQuantiles(AbstractPlotClass):
    """
    This class creates cond.quantile plots as originally proposed by Murphy, Brown and Chen (1989) [But in log scale]

    Link to paper: https://journals.ametsoc.org/doi/pdf/10.1175/1520-0434%281989%29004%3C0485%3ADVOTF%3E2.0.CO%3B2
    """
    # ignore warnings if nans appear in quantile grouping
    warnings.filterwarnings("ignore", message="All-NaN slice encountered")
    # ignore warnings if mean is calculated on nans
    warnings.filterwarnings("ignore", message="Mean of empty slice")
    # ignore warnings for y tick = 0 on log scale (instead of 0.00001 or similar)
    warnings.filterwarnings("ignore", message="Attempted to set non-positive bottom ylim on a log-scaled axis.")

    def __init__(self, stations: List, data_pred_path: str, plot_folder: str = ".", plot_per_seasons=True,
                 rolling_window: int = 3, model_mame: str = "CNN", obs_name: str = "obs", **kwargs):
        """

        :param stations: all stations to plot
        :param data_pred_path: path to dir which contains the forecasts as .nc files
        :param plot_folder: path where the plots are stored
        :param plot_per_seasons: if `True' create cond. quantile plots for seasons (DJF, MAM, JJA, SON) individually
        :param rolling_window: smoothing of quantiles (3 is used by Murphy et al.)
        :param model_mame: name of the model prediction as stored in netCDF file (for example "CNN")
        :param obs_name: name of observation as stored in netCDF file (for example "obs")
        :param kwargs: Some further arguments which are listed in self._opts
        """
        super().__init__(plot_folder, "conditional_quantiles")

        self._data_pred_path = data_pred_path
        self._stations = stations
        self._rolling_window = rolling_window
        self._model_name = model_mame
        self._obs_name = obs_name

        self._opts = {"q": kwargs.get("q", [.1, .25, .5, .75, .9]),
                      "linetype": kwargs.get("linetype", [':', '-.', '--', '-.', ':']),
                      "legend": kwargs.get("legend",
                                           ['.10th and .90th quantile', '.25th and .75th quantile', '.50th quantile',
                                            'reference 1:1']),
                      "data_unit": kwargs.get("data_unit", "ppb"),
                      }
        if plot_per_seasons is True:
            self.seasons = ['DJF', 'MAM', 'JJA', 'SON']
        else:
            self.seasons = ""
        self._data = self._load_data()
        self._bins = self._get_bins_from_rage_of_data()

        self._plot()

    def _load_data(self):
        """
        This method loads forcast data

        :return:
        """
        logging.debug("... load data")
        data_collector = []
        for station in self._stations:
            file = os.path.join(self._data_pred_path, f"forecasts_{station}_test.nc")
            data_tmp = xr.open_dataarray(file)
            data_collector.append(data_tmp.loc[:, :, [self._model_name, self._obs_name]].assign_coords(station=station))
        res = xr.concat(data_collector, dim='station').transpose('index', 'type', 'ahead', 'station')
        return res

    def _segment_data(self, data, x_model):
        """
        This method creates segmented data which is used for cond. quantile plots

        :param data:
        :param x_model:
        :return:
        """
        logging.debug("... segment data")
        # combine index and station to multi index
        data = data.stack(z=['index', 'station'])
        # replace multi index by simple position index (order is not relevant anymore)
        data.coords['z'] = range(len(data.coords['z']))
        # segment data of x_model into bins
        data.loc[x_model, ...] = data.loc[x_model, ...].to_pandas().T.apply(pd.cut, bins=self._bins,
                                                                                labels=self._bins[1:]).T.values
        return data

    @staticmethod
    def _labels(plot_type, data_unit="ppb"):
        """
        Helper method to correctly assign (x,y) labels to plots, depending on like-base or cali-ref factorization

        :param plot_type:
        :param data_unit:
        :return:
        """
        names = (f"forecast concentration (in {data_unit})", f"observed concentration (in {data_unit})")
        if plot_type == "obs":
            return names
        else:
            return names[::-1]

    def _get_bins_from_rage_of_data(self):
        """
        Get array of bins to use for quantiles

        :return:
        """
        return np.arange(0, math.ceil(self._data.max().max()) + 1, 1).astype(int)

    def _create_quantile_panel(self, data, x_model, y_model):
        """
        Clculate quantiles

        :param data:
        :param x_model:
        :param y_model:
        :return:
        """
        logging.debug("... create quantile panel")
        # create empty xarray with dims: time steps ahead, quantiles, bin index (numbers create in previous step)
        quantile_panel = xr.DataArray(
            np.full([data.ahead.shape[0], len(self._opts["q"]), self._bins[1:].shape[0]], np.nan),
            coords=[data.ahead, self._opts["q"], self._bins[1:]], dims=['ahead', 'quantiles', 'categories'])
        # ensure that the coordinates are in the right order
        quantile_panel = quantile_panel.transpose('ahead', 'quantiles', 'categories')
        # calculate for each bin of the pred_name data the quantiles of the ref_name data
        for bin in self._bins[1:]:
            mask = (data.loc[x_model, ...] == bin)
            quantile_panel.loc[..., bin] = data.loc[y_model, ...].where(mask).quantile(self._opts["q"],
                                                                                                    dim=['z']).T
        return quantile_panel

    @staticmethod
    def add_affix(x):
        """
        Helper method to add additional information on plot name

        :param x:
        :return:
        """
        return f"_{x}" if len(x) > 0 else ""

    def _prepare_plots(self, data, x_model, y_model):
        """
        Get segmented_data and quantile_panel

        :param data:
        :param x_model:
        :param y_model:
        :return:
        """
        segmented_data = self._segment_data(data, x_model)
        quantile_panel = self._create_quantile_panel(segmented_data, x_model, y_model)
        return segmented_data, quantile_panel

    def _plot(self):
        """
        Main plot call

        :return:
        """
        logging.info(f"start plotting {self.__class__.__name__}, scheduled number of plots: {(len(self.seasons) + 1) * 2}")

        if len(self.seasons) > 0:
            self._plot_seasons()
        self._plot_all()

    def _plot_seasons(self):
        """
        Seasonal plot call

        :return:
        """
        for season in self.seasons:
            self._plot_base(data=self._data.where(self._data['index.season'] == season), x_model=self._model_name,
                            y_model=self._obs_name, plot_name_affix="cali-ref", season=season)
            self._plot_base(data=self._data.where(self._data['index.season'] == season), x_model=self._obs_name,
                            y_model=self._model_name, plot_name_affix="like-base", season=season)

    def _plot_all(self):
        """
        Full plot call

        :return:
        """
        self._plot_base(data=self._data, x_model=self._model_name, y_model=self._obs_name, plot_name_affix="cali-ref")
        self._plot_base(data=self._data, x_model=self._obs_name, y_model=self._model_name, plot_name_affix="like-base")

    @TimeTrackingWrapper
    def _plot_base(self, data, x_model, y_model, plot_name_affix, season=""):
        """
        Base method to create cond. quantile plots. Is called from _plot_all and _plot_seasonal

        :param data: data which is used to create cond. quantile plot
        :param x_model: name of model on x axis (can also be obs)
        :param y_model: name of model on y axis (can also be obs)
        :param plot_name_affix: should be `cali-ref' or `like-base'
        :param season: List of seasons to use
        :return:
        """

        segmented_data, quantile_panel = self._prepare_plots(data, x_model, y_model)
        ylabel, xlabel = self._labels(x_model, self._opts["data_unit"])
        plot_name = f"{self.plot_name}{self.add_affix(season)}{self.add_affix(plot_name_affix)}_plot.pdf"
        #f"{base_name}{add_affix(season)}{add_affix(plot_name_affix)}_plot.pdf"
        plot_path = os.path.join(os.path.abspath(self.plot_folder), plot_name)
        pdf_pages = matplotlib.backends.backend_pdf.PdfPages(plot_path)
        logging.debug(f"... plot path is {plot_path}")

        # create plot for each time step ahead
        y2_max = 0
        for iteration, d in enumerate(segmented_data.ahead):
            logging.debug(f"... plotting {d.values} time step(s) ahead")
            # plot smoothed lines with rolling mean
            smooth_data = quantile_panel.loc[d, ...].rolling(categories=self._rolling_window,
                                                             center=True).mean().to_pandas().T
            ax = smooth_data.plot(style=self._opts["linetype"], color='black', legend=False)
            ax2 = ax.twinx()
            # add reference line
            ax.plot([0, self._bins.max()], [0, self._bins.max()], color='k', label='reference 1:1', linewidth=.8)
            # add histogram of the segmented data (pred_name)
            handles, labels = ax.get_legend_handles_labels()
            segmented_data.loc[x_model, d, :].to_pandas().hist(bins=self._bins, ax=ax2, color='k', alpha=.3, grid=False,
                                                                 rwidth=1)
            # add legend
            plt.legend(handles[:3] + [handles[-1]], self._opts["legend"], loc='upper left', fontsize='large')
            # adjust limits and set labels
            ax.set(xlim=(0, self._bins.max()), ylim=(0, self._bins.max()))
            ax.set_xlabel(xlabel, fontsize='x-large')
            ax.tick_params(axis='x', which='major', labelsize=15)
            ax.set_ylabel(ylabel, fontsize='x-large')
            ax.tick_params(axis='y', which='major', labelsize=15)
            ax2.yaxis.label.set_color('gray')
            ax2.tick_params(axis='y', colors='gray')
            ax2.yaxis.labelpad = -15
            ax2.set_yscale('log')
            if iteration == 0:
                y2_max = ax2.get_ylim()[1] + 100
            ax2.set(ylim=(0, y2_max * 10 ** 8), yticks=np.logspace(0, 4, 5))
            ax2.set_ylabel('              sample size', fontsize='x-large')
            ax2.tick_params(axis='y', which='major', labelsize=15)
            # set title and save current figure
            title = f"{d.values} time step(s) ahead{f' ({season})' if len(season) > 0 else ''}"
            plt.title(title)
            pdf_pages.savefig()
        # close all open figures / plots
        pdf_pages.close()
        plt.close('all')


@TimeTrackingWrapper
class PlotClimatologicalSkillScore(AbstractPlotClass):
    """
    Create plot of climatological skill score after Murphy (1988) as box plot over all stations. A forecast time step
    (called "ahead") is separately shown to highlight the differences for each prediction time step. Either each single
    term is plotted (score_only=False) or only the resulting scores CASE I to IV are displayed (score_only=True,
    default). Y-axis is adjusted following the data and not hard coded. The plot is saved under plot_folder path with
    name skill_score_clim_{extra_name_tag}{model_setup}.pdf and resolution of 500dpi.
    """
    def __init__(self, data: Dict, plot_folder: str = ".", score_only: bool = True, extra_name_tag: str = "",
                 model_setup: str = ""):
        """
        Sets attributes and create plot
        :param data: dictionary with station names as keys and 2D xarrays as values, consist on axis ahead and terms.
        :param plot_folder: path to save the plot (default: current directory)
        :param score_only: if true plot only scores of CASE I to IV, otherwise plot all single terms (default True)
        :param extra_name_tag: additional tag that can be included in the plot name (default "")
        :param model_setup: architecture type to specify plot name (default "CNN")
        """
        super().__init__(plot_folder, f"skill_score_clim_{extra_name_tag}{model_setup}")
        self._labels = None
        self._data = self._prepare_data(data, score_only)
        self._plot(score_only)
        self._save()

    def _prepare_data(self, data: Dict, score_only: bool) -> pd.DataFrame:
        """
        Shrink given data, if only scores are relevant. In any case, transform data to a plot friendly format. Also set
        plot labels depending on the lead time dimensions.
        :param data: dictionary with station names as keys and 2D xarrays as values
        :param score_only: if true only scores of CASE I to IV are relevant
        :return: pre-processed data set
        """
        data = helpers.dict_to_xarray(data, "station")
        self._labels = [str(i) + "d" for i in data.coords["ahead"].values]
        if score_only:
            data = data.loc[:, ["CASE I", "CASE II", "CASE III", "CASE IV"], :]
        return data.to_dataframe("data").reset_index(level=[0, 1, 2])

    def _label_add(self, score_only: bool):
        """
        Adds the phrase "terms and " if score_only is disabled or empty string (if score_only=True).
        :param score_only: if false all terms are relevant, otherwise only CASE I to IV
        :return: additional label
        """
        return "" if score_only else "terms and "

    def _plot(self, score_only):
        """
        Main plot function to plot climatological skill score.
        :param score_only: if true plot only scores of CASE I to IV, otherwise plot all single terms
        """
        fig, ax = plt.subplots()
        if not score_only:
            fig.set_size_inches(11.7, 8.27)
        sns.boxplot(x="terms", y="data", hue="ahead", data=self._data, ax=ax, whis=1., palette="Blues_d",
                    showmeans=True, meanprops={"markersize": 1, "markeredgecolor": "k"}, flierprops={"marker": "."})
        ax.axhline(y=0, color="grey", linewidth=.5)
        ax.set(ylabel=f"{self._label_add(score_only)}skill score", xlabel="", title="summary of all stations")
        handles, _ = ax.get_legend_handles_labels()
        ax.legend(handles, self._labels)
        plt.tight_layout()


@TimeTrackingWrapper
class PlotCompetitiveSkillScore(AbstractPlotClass):
    """
    Create competitive skill score for the given model setup and the reference models ordinary least squared ("ols") and
    the persistence forecast ("persi") for all lead times ("ahead"). The plot is saved under plot_folder with the name
    skill_score_competitive_{model_setup}.pdf and resolution of 500dpi.
    """
    def __init__(self, data: pd.DataFrame, plot_folder=".", model_setup="CNN"):
        """
        :param data: data frame with index=['cnn-persi', 'ols-persi', 'cnn-ols'] and columns "ahead" containing the pre-
            calculated comparisons for cnn, persistence and ols.
        :param plot_folder: path to save the plot (default: current directory)
        :param model_setup: architecture type (default "CNN")
        """
        super().__init__(plot_folder, f"skill_score_competitive_{model_setup}")
        self._labels = None
        self._data = self._prepare_data(data)
        self._plot()
        self._save()

    def _prepare_data(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Reformat given data and create plot labels. Introduces the dimensions stations and comparison
        :param data: data frame with index=['cnn-persi', 'ols-persi', 'cnn-ols'] and columns "ahead" containing the pre-
            calculated comparisons for cnn, persistence and ols.
        :return: processed data
        """
        data = pd.concat(data, axis=0)
        data = xr.DataArray(data, dims=["stations", "ahead"]).unstack("stations")
        data = data.rename({"stations_level_0": "stations", "stations_level_1": "comparison"})
        data = data.to_dataframe("data").unstack(level=1).swaplevel()
        data.columns = data.columns.levels[1]
        self._labels = [str(i) + "d" for i in data.index.levels[1].values]
        return data.stack(level=0).reset_index(level=2, drop=True).reset_index(name="data")

    def _plot(self):
        """
        Main plot function to plot skill scores of the comparisons cnn-persi, ols-persi and cnn-ols.
        """
        fig, ax = plt.subplots()
        sns.boxplot(x="comparison", y="data", hue="ahead", data=self._data, whis=1., ax=ax, palette="Blues_d",
                    showmeans=True, meanprops={"markersize": 3, "markeredgecolor": "k"}, flierprops={"marker": "."},
                    order=["cnn-persi", "ols-persi", "cnn-ols"])
        ax.axhline(y=0, color="grey", linewidth=.5)

        ax.set(ylabel="skill score", xlabel="competing models", title="summary of all stations", ylim=self._ylim())
        handles, _ = ax.get_legend_handles_labels()
        ax.legend(handles, self._labels)
        plt.tight_layout()

    def _ylim(self) -> Tuple[float, float]:
        """
        Calculate y-axis limits from data. Lower is the minimum of either 0 or data's minimum (reduced by small
        subtrahend) and upper limit is data's maximum (increased by a small addend).
        :return:
        """
        lower = np.min([0, helpers.float_round(self._data.min()[2], 2) - 0.1])
        upper = helpers.float_round(self._data.max()[2], 2) + 0.1
        return lower, upper


@TimeTrackingWrapper
class PlotBootstrapSkillScore(AbstractPlotClass):
    """
    Create plot of climatological skill score after Murphy (1988) as box plot over all stations. A forecast time step
    (called "ahead") is separately shown to highlight the differences for each prediction time step. Either each single
    term is plotted (score_only=False) or only the resulting scores CASE I to IV are displayed (score_only=True,
    default). Y-axis is adjusted following the data and not hard coded. The plot is saved under plot_folder path with
    name skill_score_clim_{extra_name_tag}{model_setup}.pdf and resolution of 500dpi.
    """

    def __init__(self, data: Dict, plot_folder: str = ".", model_setup: str = ""):
        """
        Sets attributes and create plot
        :param data: dictionary with station names as keys and 2D xarrays as values, consist on axis ahead and terms.
        :param plot_folder: path to save the plot (default: current directory)
        :param model_setup: architecture type to specify plot name (default "CNN")
        """
        super().__init__(plot_folder, f"skill_score_bootstrap_{model_setup}")
        self._labels = None
        self._x_name = "boot_var"
        self._data = self._prepare_data(data)
        self._plot()
        self._save()

    def _prepare_data(self, data: Dict) -> pd.DataFrame:
        """
        Shrink given data, if only scores are relevant. In any case, transform data to a plot friendly format. Also set
        plot labels depending on the lead time dimensions.
        :param data: dictionary with station names as keys and 2D xarrays as values
        :return: pre-processed data set
        """
        data = helpers.dict_to_xarray(data, "station").sortby(self._x_name)
        self._labels = [str(i) + "d" for i in data.coords["ahead"].values]
        return data.to_dataframe("data").reset_index(level=[0, 1, 2])

    def _label_add(self, score_only: bool):
        """
        Adds the phrase "terms and " if score_only is disabled or empty string (if score_only=True).
        :param score_only: if false all terms are relevant, otherwise only CASE I to IV
        :return: additional label
        """
        return "" if score_only else "terms and "

    def _plot(self):
        """
        Main plot function to plot climatological skill score.
        """
        fig, ax = plt.subplots()
        sns.boxplot(x=self._x_name, y="data", hue="ahead", data=self._data, ax=ax, whis=1., palette="Blues_d",
                    showmeans=True, meanprops={"markersize": 1, "markeredgecolor": "k"}, flierprops={"marker": "."})
        ax.axhline(y=0, color="grey", linewidth=.5)
        ax.set(ylabel=f"skill score", xlabel="", title="summary of all stations")
        handles, _ = ax.get_legend_handles_labels()
        ax.legend(handles, self._labels)
        plt.tight_layout()


@TimeTrackingWrapper
class PlotTimeSeries:

    def __init__(self, stations: List, data_path: str, name: str, window_lead_time: int = None, plot_folder: str = ".",
                 sampling="daily"):
        self._data_path = data_path
        self._data_name = name
        self._stations = stations
        self._window_lead_time = self._get_window_lead_time(window_lead_time)
        self._sampling = self._get_sampling(sampling)
        self._plot(plot_folder)

    @staticmethod
    def _get_sampling(sampling):
        if sampling == "daily":
            return "D"
        elif sampling == "hourly":
            return "h"

    def _get_window_lead_time(self, window_lead_time: int):
        """
        Extract the lead time from data and arguments. If window_lead_time is not given, extract this information from
        data itself by the number of ahead dimensions. If given, check if data supports the give length. If the number
        of ahead dimensions in data is lower than the given lead time, data's lead time is used.
        :param window_lead_time: lead time from arguments to validate
        :return: validated lead time, comes either from given argument or from data itself
        """
        ahead_steps = len(self._load_data(self._stations[0]).ahead)
        if window_lead_time is None:
            window_lead_time = ahead_steps
        return min(ahead_steps, window_lead_time)

    def _load_data(self, station):
        logging.debug(f"... preprocess station {station}")
        file_name = os.path.join(self._data_path, self._data_name % station)
        data = xr.open_dataarray(file_name)
        return data.sel(type=["CNN", "obs"])

    def _plot(self, plot_folder):
        pdf_pages = self._create_pdf_pages(plot_folder)
        start, end = self._get_time_range(self._load_data(self._stations[0]))
        for pos, station in enumerate(self._stations):
            data = self._load_data(station)
            fig, axes, factor = self._create_subplots(start, end)
            nan_list = []
            for i_year in range(end - start + 1):
                data_year = data.sel(index=f"{start + i_year}")
                for i_half_of_year in range(factor):
                    pos = factor * i_year + i_half_of_year
                    plot_data = self._create_plot_data(data_year, factor, i_half_of_year)
                    self._plot_obs(axes[pos], plot_data)
                    self._plot_ahead(axes[pos], plot_data)
                    if np.isnan(plot_data.values).all():
                        nan_list.append(pos)
            self._clean_up_axes(nan_list, axes, fig)
            self._save_page(station, pdf_pages)
        pdf_pages.close()
        plt.close('all')

    @staticmethod
    def _clean_up_axes(nan_list, axes, fig):
        for i in reversed(nan_list):
            fig.delaxes(axes[i])

    @staticmethod
    def _save_page(station, pdf_pages):
        plt.suptitle(station)
        plt.legend()
        plt.tight_layout()
        pdf_pages.savefig(dpi=500)

    @staticmethod
    def _create_plot_data(data, factor, running_index):
        if factor > 1:
            if running_index == 0:
                data = data.where(data["index.month"] < 7)
            else:
                data = data.where(data["index.month"] >= 7)
        return data

    def _create_subplots(self, start, end):
        factor = 1
        if self._sampling == "h":
            factor = 2
        f, ax = plt.subplots((end - start + 1) * factor, sharey=True, figsize=(50, 30))
        return f, ax, factor

    def _plot_ahead(self, ax, data):
        color = sns.color_palette("Blues_d", self._window_lead_time).as_hex()
        for ahead in data.coords["ahead"].values:
            plot_data = data.sel(type="CNN", ahead=ahead).drop(["type", "ahead"]).squeeze().shift(index=ahead)
            label = f"{ahead}{self._sampling}"
            ax.plot(plot_data, color=color[ahead-1], label=label)

    def _plot_obs(self, ax, data):
        ahead = 1
        obs_data = data.sel(type="obs", ahead=ahead).shift(index=ahead)
        # index = data.index + np.timedelta64(1, self._sampling)
        # ax.plot(index, obs_data.values, color=matplotlib.colors.cnames["green"], label="obs")
        ax.plot(obs_data, color=matplotlib.colors.cnames["green"], label="obs")

    @staticmethod
    def _get_time_range(data):
        def f(x, f_x):
            return pd.to_datetime(f_x(x.index.values)).year
        return f(data, min), f(data, max)

    @staticmethod
    def _create_pdf_pages(plot_folder):
        """
        Standard save method to store plot locally. The name of this plot is static.
        :param plot_folder: path to save the plot
        """
        plot_name = os.path.join(os.path.abspath(plot_folder), 'timeseries_plot.pdf')
        logging.debug(f"... save plot to {plot_name}")
        return matplotlib.backends.backend_pdf.PdfPages(plot_name)


@TimeTrackingWrapper
class PlotAvailability(AbstractPlotClass):

    def __init__(self, generators: Dict[str, DataGenerator], plot_folder: str = ".", sampling="daily",
                 summary_name="data availability"):
        # create standard Gantt plot for all stations (currently in single pdf file with single page)
        super().__init__(plot_folder, "data_availability")
        self.sampling = self._get_sampling(sampling)
        plot_dict = self._prepare_data(generators)
        lgd = self._plot(plot_dict)
        self._save(bbox_extra_artists=(lgd, ), bbox_inches="tight")
        # create summary Gantt plot (is data in at least one station available)
        self.plot_name += "_summary"
        plot_dict_summary = self._summarise_data(generators, summary_name)
        lgd = self._plot(plot_dict_summary)
        self._save(bbox_extra_artists=(lgd, ), bbox_inches="tight")
        # combination of station and summary plot, last element is summary broken bar
        self.plot_name = "data_availability_combined"
        plot_dict_summary.update(plot_dict)
        lgd = self._plot(plot_dict_summary)
        self._save(bbox_extra_artists=(lgd, ), bbox_inches="tight")

    @staticmethod
    def _get_sampling(sampling):
        if sampling == "daily":
            return "D"
        elif sampling == "hourly":
            return "h"

    def _prepare_data(self, generators: Dict[str, DataGenerator]):
        plt_dict = {}
        for subset, generator in generators.items():
            stations = generator.stations
            for station in stations:
                station_data = generator.get_data_generator(station)
                labels = station_data.get_transposed_label().resample(datetime=self.sampling, skipna=True).mean()
                labels_bool = labels.sel(window=1).notnull()
                group = (labels_bool != labels_bool.shift(datetime=1)).cumsum()
                plot_data = pd.DataFrame({"avail": labels_bool.values, "group": group.values}, index=labels.datetime.values)
                t = plot_data.groupby("group").apply(lambda x: (x["avail"].head(1)[0], x.index[0], x.shape[0]))
                t2 = [i[1:] for i in t if i[0]]

                if plt_dict.get(station) is None:
                    plt_dict[station] = {subset: t2}
                else:
                    plt_dict[station].update({subset: t2})
        return plt_dict

    def _summarise_data(self, generators: Dict[str, DataGenerator], summary_name: str):
        plt_dict = {}
        for subset, generator in generators.items():
            all_data = None
            stations = generator.stations
            for station in stations:
                station_data = generator.get_data_generator(station)
                labels = station_data.get_transposed_label().resample(datetime=self.sampling, skipna=True).mean()
                labels_bool = labels.sel(window=1).notnull()
                if all_data is None:
                    all_data = labels_bool
                else:
                    tmp = all_data.combine_first(labels_bool)  # expand dims to merged datetime coords
                    all_data = np.logical_or(tmp, labels_bool).combine_first(all_data)  # apply logical on merge and fill missing with all_data

            group = (all_data != all_data.shift(datetime=1)).cumsum()
            plot_data = pd.DataFrame({"avail": all_data.values, "group": group.values}, index=all_data.datetime.values)
            t = plot_data.groupby("group").apply(lambda x: (x["avail"].head(1)[0], x.index[0], x.shape[0]))
            t2 = [i[1:] for i in t if i[0]]
            if plt_dict.get(summary_name) is None:
                plt_dict[summary_name] = {subset: t2}
            else:
                plt_dict[summary_name].update({subset: t2})
        return plt_dict

    def _plot(self, plt_dict):
        # colors = {"train": "orange", "val": "blueishgreen", "test": "skyblue"}  # color names
        colors = {"train": "#e69f00", "val": "#009e73", "test": "#56b4e9"}  # hex code
        # colors = {"train": (230, 159, 0), "val": (0, 158, 115), "test": (86, 180, 233)}  # in rgb but as abs values
        pos = 0
        height = 0.8  # should be <= 1
        yticklabels = []
        number_of_stations = len(plt_dict.keys())
        fig, ax = plt.subplots(figsize=(10, number_of_stations/3))
        for station, d in sorted(plt_dict.items(), reverse=True):
            pos += 1
            for subset, color in colors.items():
                plt_data = d.get(subset)
                if plt_data is None:
                    continue
                ax.broken_barh(plt_data, (pos, height), color=color, edgecolor="white")
            yticklabels.append(station)

        ax.set_ylim([height, number_of_stations + 1])
        ax.set_yticks(np.arange(len(plt_dict.keys()))+1+height/2)
        ax.set_yticklabels(yticklabels)
        handles = [mpatches.Patch(color=c, label=k) for k, c in colors.items()]
        lgd = plt.legend(handles=handles, bbox_to_anchor=(0, 1, 1, 0.2), loc="lower center", ncol=len(handles))
        return lgd


if __name__ == "__main__":
    stations = ['DEBW107', 'DEBY081', 'DEBW013', 'DEBW076', 'DEBW087']
    path = "../../testrun_network/forecasts"
    plt_path = "../../"

    con_quan_cls = PlotConditionalQuantiles(stations, path, plt_path)

